// Stage3D Shoot-em-up Tutorial Part 5
// by Christer Kaitila - www.mcfunkypants.com

// GameLevels.as
// This class parses .CSV level data strings
// that define the locations of tiles from a spritesheet
// Example levels were created using the OGMO editor,
// but could be designed by hand or any number of other
// freeware game level editors that can output .csv
// This can be a .txt, .csv, .oel, .etc file
// - we will strip all xml/html tags (if any) 
// - we only care about raw csv data
// Our game can access the current level with:
// spriteId = myLevel.data[x][y];

package
{
	import flash.display3D.Context3DProgramType;
	public class GameLevels
	{
		// v5 the farthest column in the level
		// used to detect when the map is complete
		public var levelLength:int = 0;

		// the "demo" level seen during the title screen
		[Embed(source = '../assets/level0.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL0:Class;
		private var level0data:String = new LEVEL0;
		[Embed(source = '../assets/terrain0.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL0TERRAIN:Class;
		private var level0terrain:String = new LEVEL0TERRAIN;

		// level 1
		[Embed(source = '../assets/level1.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL1:Class;
		private var level1data:String = new LEVEL1;
		[Embed(source = '../assets/terrain1.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL1TERRAIN:Class;
		private var level1terrain:String = new LEVEL1TERRAIN;

		// level 2
		[Embed(source = '../assets/level2.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL2:Class;
		private var level2data:String = new LEVEL2;
		[Embed(source = '../assets/terrain2.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL2TERRAIN:Class;
		private var level2terrain:String = new LEVEL2TERRAIN;

		// level 3
		[Embed(source = '../assets/level3.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL3:Class;
		private var level3data:String = new LEVEL3;
		[Embed(source = '../assets/terrain3.oel', mimeType = 'application/octet-stream')] 
		private static const LEVEL3TERRAIN:Class;
		private var level3terrain:String = new LEVEL3TERRAIN;

		// the currently loaded level data
		public var data:Array = [];
		
		public function GameLevels()
		{
		}
		
		private function stripTags(str:String):String
		{
			var pattern:RegExp = /<\/?[a-zA-Z0-9]+.*?>/gim;
			return str.replace(pattern, "");
		}
	
		private function parseLevelData(lvl:String):Array
		{
			var levelString:String;
			var temps:Array;
			var nextValue:int;
			var output:Array = [];
			var nextrow:int;

			// how many columns wide is the map?
			// note: some rows may be shorter
			levelLength = 0;
			
			switch (lvl)
			{
				case "level0" : levelString = stripTags(level0data); break;
				case "terrain0" : levelString = stripTags(level0terrain); break;
				case "level1" : levelString = stripTags(level1data); break;
				case "terrain1" : levelString = stripTags(level1terrain); break;
				case "level2" : levelString = stripTags(level2data); break;
				case "terrain2" : levelString = stripTags(level2terrain); break;
				case "level3" : levelString = stripTags(level3data); break;
				case "terrain3" : levelString = stripTags(level3terrain); break;
				default:
					return output; 
			}
			
			//trace("Level " + num + " data:\n" + levelString);
			var lines:Array = levelString.split(/\r\n|\n|\r/);
			for (var row:int = 0; row < lines.length; row++)
			{
				// split the string by comma
				temps = lines[row].split(",");
				if (temps.length > 1)
				{
					nextrow = output.push([]) - 1;
					// turn the string values into integers
					for (var col:int = 0; col < temps.length; col++)
					{
						if (temps[col] == "") temps[col] = "-1";
						nextValue = parseInt(temps[col]);
						if (nextValue < 0) nextValue = -1; // we still need blanks
						// trace('row '+ nextrow + ' nextValue=' + nextValue);
						
						// v5 remember longest column so we know when we've reached the end
						if (col > levelLength) levelLength = col;
						
						output[nextrow].push(nextValue);
						
					}
					//trace('Level row '+nextrow+':\n' + String(output[nextrow]));
				}
			}
			//trace('Level output data:\n' + String(output));
			return output;
		}
		
		public function loadLevel(lvl:String):void
		{
			trace("Loading level " + lvl);
			data = parseLevelData(lvl);
		}
		
	} // end class
} // end package