// Stage3D Shoot-em-up Tutorial Part 5
// by Christer Kaitila - www.mcfunkypants.com
// Created for active.tutsplus.com

package 
{
	[SWF(width = "600", height = "400", frameRate = "60", backgroundColor = "#000000")]

	import flash.display3D.*;
	import flash.display.Sprite;
	import flash.display.StageAlign;
	import flash.display.StageQuality;
	import flash.display.StageScaleMode;
	import flash.events.Event;
	import flash.events.ErrorEvent;
	import flash.events.MouseEvent;
	import flash.geom.Rectangle;
	import flash.utils.getTimer;
	import flash.geom.Point; // v5
		
	public class Main extends Sprite 
	{
		// the game save/load system
		public var saved:GameSaves; // v5
		
		// the entity spritesheet (ships, particles)
		[Embed(source="../assets/sprites.png")]
		private var EntitySourceImage : Class;

		// the terrain spritesheet
		[Embed(source="../assets/terrain.png")]
		private var TerrainSourceImage : Class;
		
		// the keyboard control system
		private var _controls : GameControls;
		// don't update the menu too fast
		private var nothingPressedLastFrame:Boolean = false;
		// timestamp of the current frame
		public var currentTime:int;
		// for framerate independent speeds
		public var currentFrameMs:int;
		public var previousFrameTime:int;
		
		// player one's entity
		public var thePlayer:Entity;
		// movement speed in pixels per second
		public var playerSpeed:Number = 128;
		// timestamp when next shot can be fired
		private var nextFireTime:uint = 0;
		// how many ms between shots
		private var fireDelay:uint = 200;
		
		// main menu = 0 or current level number
		private var _state:int = 0;
		// the title screen batch
		private var _mainmenu:GameMenu;
		// the sound system
		private var _sfx:GameSound;	
		// the background stars
		private var _bg:GameBackground;	
		
		private var _terrain:EntityManager;
		private var _entities:EntityManager;
		private var _spriteStage:LiteSpriteStage;
		private var _gui:GameGUI;
		private var _width:Number = 600;
		private var _height:Number = 400;
		public var context3D:Context3D;
		
		// constructor function for our game
		public function Main():void 
		{
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		// called once flash is ready
		private function init(e:Event = null):void 
		{
			_controls = new GameControls(stage);
			removeEventListener(Event.ADDED_TO_STAGE, init);
			stage.quality = StageQuality.LOW;
			stage.align = StageAlign.TOP_LEFT;
			stage.scaleMode = StageScaleMode.NO_SCALE;
			stage.addEventListener(Event.RESIZE, onResizeEvent);
			trace("Init Stage3D...");
			_gui = new GameGUI("");
			addChild(_gui);
			stage.stage3Ds[0].addEventListener(Event.CONTEXT3D_CREATE, onContext3DCreate);
			stage.stage3Ds[0].addEventListener(ErrorEvent.ERROR, errorHandler);
			stage.stage3Ds[0].requestContext3D(Context3DRenderMode.AUTO);
			trace("Stage3D requested...");		
			_sfx = new GameSound();
		}
				
		// this is called when the 3d card has been set up
		// and is ready for rendering using stage3d
		private function onContext3DCreate(e:Event):void 
		{
			trace("Stage3D context created! Init sprite engine...");
			context3D = stage.stage3Ds[0].context3D;
			initSpriteEngine();
		}
		
		// this can be called when using an old version of flash
		// or if the html does not include wmode=direct
		private function errorHandler(e:ErrorEvent):void 
		{
			trace("Error while setting up Stage3D: "+e.errorID+" - " +e.text);
		}

		protected function onResizeEvent(event:Event) : void
		{
			trace("resize event...");
			
			// Set correct dimensions if we resize
			_width = stage.stageWidth;
			_height = stage.stageHeight;
			
			// Resize Stage3D to continue to fit screen
			var view:Rectangle = new Rectangle(0, 0, _width, _height);
			if ( _spriteStage != null ) {
				_spriteStage.position = view;
			}
			if(_terrain != null) {
				_terrain.setPosition(view);
			}
			if(_entities != null) {
				_entities.setPosition(view);
			}
			if(_mainmenu != null) {
				_mainmenu.setPosition(view);
			}
		}
		
		private function initSpriteEngine():void 
		{
			// init a gpu sprite system
			//var view:Rectangle = new Rectangle(0,0,_width,_height)
			var stageRect:Rectangle = new Rectangle(0, 0, _width, _height); 
			_spriteStage = new LiteSpriteStage(stage.stage3Ds[0], context3D, stageRect);
			_spriteStage.configureBackBuffer(_width,_height);
			
			// create the background stars
			trace("Init background...");
			_bg = new GameBackground(stageRect);
			_bg.createBatch(context3D);
			_spriteStage.addBatch(_bg.batch);
			_bg.initBackground();
			
			// create the terrain spritesheet and batch
			trace("Init Terrain...");
			_terrain = new EntityManager(stageRect);
			_terrain.SourceImage = TerrainSourceImage;
			_terrain.SpritesPerRow = 16;
			_terrain.SpritesPerCol = 16;
			_terrain.defaultSpeed = 90;
			_terrain.defaultScale = 1.5;
			_terrain.levelTilesize = 48; 
			_terrain.createBatch(context3D, 0.001); // a little UV padding required
			_spriteStage.addBatch(_terrain.batch);
			_terrain.changeLevels('terrain' + _state);

			// create a single rendering batch
			// which will draw all sprites in one pass
			trace("Init Entities...");
			_entities = new EntityManager(stageRect);
			_entities.SourceImage = EntitySourceImage;
			_entities.defaultScale = 1.5; // 1
			_entities.levelTilesize = 48; 
			_entities.createBatch(context3D);
			_entities.sfx = _sfx;
			_spriteStage.addBatch(_entities.batch);
			_entities.changeLevels('level' + _state);
			_entities.streamLevelEntities(true); // spawn first row of the level immediately
			
			// create the logo/titlescreen main menu
			_mainmenu = new GameMenu(stageRect);
			_mainmenu.createBatch(context3D);
			_spriteStage.addBatch(_mainmenu.batch);
			
			// tell the gui where to grab statistics from
			_gui.statsTarget = _entities; 
			
			// start the render loop
			stage.addEventListener(Event.ENTER_FRAME,onEnterFrame);

			// only used for the menu
			stage.addEventListener(MouseEvent.MOUSE_DOWN, mouseDown);   
			stage.addEventListener(MouseEvent.MOUSE_MOVE, mouseMove); 

			// set up the savegame system
			saved = new GameSaves();
			_gui.highScore = saved.score;
			_gui.level = saved.level;
		}
		
		// check player transition state (deaths, game over, etc)
		private var currentTransitionSeconds:Number = 0;
		private function handleTransitions(seconds:Number):void
		{
			// are we at a pending transition (death or level change)?
			if (thePlayer.transitionTimeLeft > 0)
			{
				currentTransitionSeconds += seconds;
				
				thePlayer.transitionTimeLeft -= seconds;
				
				if (thePlayer.transitionTimeLeft > 0)
				{	//was it a level change?
					if (thePlayer.level != _state)
					{
						if (_state == -1)
						{
							_gui.transitionText = "\n\n\n\n\n\nCONGRATULATIONS\n\n" +
								"You fought bravely and defended\n" +
								"the universe from certain doom.\n\nYou got to level " + 
								thePlayer.level + "\nwith " + thePlayer.score + " points." + 
								"\n\nCREDITS:\n\nProgramming: McFunkypants\n(mcfunkypants.com)\n\n" +
								"Art: Daniel Cook\n(lostgarden.com)\n\n" +
								"Music: MaF\n(maf464.com)\n\n" +
								"Thanks for playing!";
							_gui.transitionTf.scrollRect = new Rectangle(0,currentTransitionSeconds * 40,600,160);
						}
						else if (_state == 0)
							_gui.transitionText = "GAME OVER\nYou got to level " + thePlayer.level 
								+ "\nwith " + thePlayer.score + " points.";
						else if (_state > 1)
							_gui.transitionText = "\nLEVEL " + (_state-1) + " COMPLETE!";
						else
							_gui.transitionText = "\nLEVEL " + _state; 
					}
					else // must be a death or start of a map
					{
						_gui.transitionText = "Your ship was destroyed.\n\nYou have " 
							+ thePlayer.lives + (thePlayer.lives != 1 ? " lives" : " life") + " left.";
					}
					if (thePlayer.lives < 0 || thePlayer.health <= 0)
					{
						// during the death transition, spawn tons of explosions just for fun
						if (_entities.fastRandom() < 0.2)
						{
							var explosionPos:Point = new Point();
							explosionPos.x = thePlayer.sprite.position.x + _entities.fastRandom() * 128 - 64; 
							explosionPos.y = thePlayer.sprite.position.y + _entities.fastRandom() * 128 - 64; 
							_entities.particles.addExplosion(explosionPos);
						}
					}
				}
				else // transition time has elapsed
				{
					currentTransitionSeconds = 0;
					
					thePlayer.transitionTimeLeft = 0;
					
					if (_state == -1) _state = 0;
					_gui.transitionTf.scrollRect = new Rectangle(0,0,600,160);
					_gui.transitionText = "";
					
					if ((thePlayer.health <= 0) && (_state != 0)) // we died
					{
						trace("Death transition over. Respawning player.");
						thePlayer.sprite.position.y = _entities.midpoint;
						thePlayer.sprite.position.x = 64;
						thePlayer.health = 100;
						// start the level again
						_entities.changeLevels('level' + _state);
						_terrain.changeLevels('terrain' + _state);
					}
					if (thePlayer.level != _state)
					{
						trace('Level transition over. Starting level ' + _state);
						thePlayer.level = _state;
						if (_state > 1) // no need to reload at startGame
						{
							_entities.changeLevels('level' + _state);
							_terrain.changeLevels('terrain' + _state);
						}
						if (_state == 0) // game over
						{
							thePlayer.health = 100;
							thePlayer.lives = 3;
							thePlayer.sprite.visible = false;
							_entities.theOrb.sprite.visible = false;
							_spriteStage.addBatch(_mainmenu.batch);
							_entities.changeLevels('level' + _state);
							_terrain.changeLevels('terrain' + _state);
						}
					}
				}
			}
		}
		
		// run every frame by the entity manager as the player ai function
		public function playerLogic(seconds:Number):void
		{
			thePlayer.age += seconds;
			handleTransitions(seconds);
			thePlayer.speedY = thePlayer.speedX = 0;
			
			if (_state == 0) return;
			
			if (_controls.pressing.up)
				thePlayer.speedY = -playerSpeed;
			if (_controls.pressing.down)
				thePlayer.speedY = playerSpeed;
			if (_controls.pressing.left)
				thePlayer.speedX = -playerSpeed;
			if (_controls.pressing.right)
				thePlayer.speedX = playerSpeed;
				
			// v5
			if (_controls.pressing.fire && (thePlayer.health > 0))
			{
				// is it time to fire again?
				if (currentTime >= nextFireTime)
				{
					//trace("Fire!");
					nextFireTime = currentTime + fireDelay;
					_sfx.playGun(1);
					_entities.shootBullet(3);
				}
			}
				
			// keep on screen
			if (thePlayer.sprite.position.x < 0)
				thePlayer.sprite.position.x = 0;
			if (thePlayer.sprite.position.x > _width)
				thePlayer.sprite.position.x = _width;
			if (thePlayer.sprite.position.y < 0)
				thePlayer.sprite.position.y = 0;
			if (thePlayer.sprite.position.y > _height)
				thePlayer.sprite.position.y = _height;
				
			// leave a trail of particles
			_entities.particles.addParticle(63, 
				thePlayer.sprite.position.x - 12, 
				thePlayer.sprite.position.y + 2, 
				0.75, -200, 0, 0.4, NaN, NaN, -1, -1.5);
				
			// v5 if we are about to die, spew sparks as a warning
			if (thePlayer.health < 10)
				_entities.particles.addSparks(thePlayer.sprite.position, 1, 2);	
				
			// when the player gets damaged, they become
			// invulnerable for a short perod of time
			if (thePlayer.invulnerabilityTimeLeft > 0)
			{
				thePlayer.invulnerabilityTimeLeft -= seconds;
				if (thePlayer.invulnerabilityTimeLeft <= 0)
				{
					trace("Invulnerability wore off.");
					thePlayer.sprite.alpha = 1;
				}
				else // while invulnerable, flicker
				{
					thePlayer.sprite.alpha = Math.sin(thePlayer.age * 30) / Math.PI  + 0.25;
				}
			}
		}

		private function mouseDown(e:MouseEvent):void   
		{   
			trace('mouseDown at '+e.stageX+','+e.stageY);
			if (_state == 0) // are we at the main menu?
			{
				if (_mainmenu && _mainmenu.activateCurrentMenuItem(getTimer()))
				{ // if the above returns true we should start the game
					startGame();
				}
			}
		}

		private function mouseMove(e:MouseEvent):void   
		{
			if (_state == 0) // are we at the main menu?
			{
				// select menu items via mouse
				if (_mainmenu) _mainmenu.mouseHighlight(e.stageX, e.stageY);
			}
		}

		// handle any player input
		private function processInput():void
		{
			if (_state == 0) // are we at the main menu?
			{
				// select menu items via keyboard
				if (_controls.pressing.down || _controls.pressing.right)
				{
					if (nothingPressedLastFrame) 
					{
						_sfx.playGun(1);
						_mainmenu.nextMenuItem();
						nothingPressedLastFrame = false;
					}
				}
				else if (_controls.pressing.up || _controls.pressing.left)
				{
					if (nothingPressedLastFrame) 
					{
						_sfx.playGun(1);
						_mainmenu.prevMenuItem();
						nothingPressedLastFrame = false;
					}
				}
				else if (_controls.pressing.fire)
				{
					if (_mainmenu.activateCurrentMenuItem(getTimer()))
					{ // if the above returns true we should start the game
						startGame();
					}
				}
				else
				{
					// this ensures the menu doesn't change too fast
					nothingPressedLastFrame = true;
				}
			}
			// v5 - player firing moved to playerLogic function
		}
		
		private function startGame():void
		{
			trace("Starting game!");
			_state = 1;
			_spriteStage.removeBatch(_mainmenu.batch);
			_sfx.playMusic();
			
			// add the player entity to the game!
			if (!thePlayer) 
				thePlayer = _entities.addPlayer(playerLogic);
			else // on subsequent games // v5
				thePlayer.sprite.visible = true;
			if (_entities.theOrb)
				_entities.theOrb.sprite.visible = true;
				
			// load level one (and clear demo entities)
			_entities.changeLevels('level' + _state);
			_terrain.changeLevels('terrain' + _state);
			
			// reset the player position
			thePlayer.level = 0; // it will transition to 1
			thePlayer.score = 0;
			thePlayer.lives = 3;
			thePlayer.sprite.position.x = 64;
			thePlayer.sprite.position.y = _entities.midpoint;

			// add a "welcome message"
			thePlayer.transitionTimeLeft = thePlayer.transitionSeconds;
			
			// make the player invulnerable at first
			thePlayer.invulnerabilityTimeLeft = thePlayer.transitionSeconds + thePlayer.invulnerabilitySecsWhenHit;
			
		}
		
		// v5 triggered if the player loses all lives
		private function gameOver():void
		{
			trace("================ GAME OVER ================");

			// save game
			if (saved.level < thePlayer.level)
				saved.level = thePlayer.level;
			if (saved.score < thePlayer.score)
			{
				saved.score = thePlayer.score;
				_gui.highScore = thePlayer.score;
			}

			_state = 0;

			thePlayer.transitionTimeLeft = thePlayer.transitionSeconds;

		}
		
		// v5 detect if we just died, etc.
		private function checkPlayerState():void
		{
			if (_state == 0) return;
			if (thePlayer)
			{
				if (thePlayer.lives < 0)
				{
					gameOver();
				}
			}
		}
		
		// v5 check to see if we reached the end of the map/game
		private function checkMapState():void
		{
			// main menu or gameover credits?
			if (_state < 1) return;
			// already transitioning?
			if (thePlayer.level != _state) return;
			
			// allow some extra spaces for the level to scroll past
			// the player and then call the level complete.
			if (_terrain.levelPrevCol > _terrain.level.levelLength + 16)
			{
				trace("LEVEL " + _state  + " COMPLETED!");
				
				_state++;
				
				thePlayer.transitionTimeLeft = thePlayer.transitionSeconds;
				
				if (_entities.level.levelLength == 0)
				{
					trace("NO MORE LEVELS REMAIN! GAME OVER!");
					rollTheCredits();
				}
			}
		}
		
		// display the "game cleared" screen
		private function rollTheCredits():void
		{
			gameOver();
			_state = -1;
			thePlayer.transitionTimeLeft = thePlayer.transitionSeconds * 3;
		}
		
		// this function draws the scene every frame
		private function onEnterFrame(e:Event):void 
		{
			try 
			{
				// grab timestamp of current frame
				currentTime = getTimer();
				currentFrameMs = currentTime - previousFrameTime;
				previousFrameTime = currentTime;
				
				// erase the previous frame
				context3D.clear(0, 0, 0, 1);
				
				// for debugging the input manager, update the gui
				// _gui.titleText = _controls.textDescription();
				
				// process any player input
				processInput();

				// scroll the background
				if (_entities.thePlayer) _bg.yParallax(_entities.thePlayer.sprite.position.y / _height);
				_bg.update(currentTime);
				
				// update the main menu titlescreen
				if (_state == 0)
					_mainmenu.update(currentTime);
				
				// move/animate all entities
				_terrain.update(currentFrameMs);
				_entities.update(currentFrameMs);
				
				// keep adding more sprites - IF we need to
				_terrain.streamLevelEntities(false);
				_entities.streamLevelEntities(true);
				
				// draw all entities
				_spriteStage.render();

				// update the screen
				context3D.present();
				
				// check for gameover/death
				checkPlayerState();
				
				// check for the end of the level
				checkMapState();
			}
			catch (e:Error) 
			{
				// this can happen if the computer goes to sleep and
				// then re-awakens, requiring reinitialization of stage3D
				// (the onContext3DCreate will fire again)
			}
		}
	} // end class
} // end package